'''Crude audiobook support for MPD (Music Player Daemon).'''

import argparse
import dataclasses
import os
import datetime
import fileinput
import logging
import typing

import musicpd

__version__ = '0.1.0a1'

_logger = logging.getLogger(__name__)

StickerMap = typing.Dict[str, str]


class MpdabException(Exception):
    '''Base exception raised by mpdab.'''


class InvalidItemException(MpdabException):
    '''Raised when an Item cannot be created from provided stickers.'''


@dataclasses.dataclass(order=True)
class Item:
    uri: str = dataclasses.field(compare=False)
    audiobook_name: str
    sequence: int
    last_position: typing.Optional[float] = dataclasses.field(
        default=None, compare=False)
    last_played: typing.Optional[datetime.datetime] = dataclasses.field(
        default=None, compare=False)


def parse_stickers(stickers: typing.List[str]) -> StickerMap:
    sticker_map = {}
    for sticker in stickers:
        k, v = sticker.split('=', 1)
        sticker_map[k] = v
    return sticker_map


def get_client_from_args(args: argparse.Namespace) -> musicpd.MPDClient:
    pwd, host = None, args.host
    if '@' in host:
        pwd, host = host.split('@', 1)

    client = musicpd.MPDClient()
    client.pwd = pwd
    client.host = host
    client.port = args.port
    return client


def get_item_from_sticker_map(song_uri: str, stickers: StickerMap) -> Item:
    if not all((
        stickers,
        stickers.get('audiobook'),
        stickers.get('sequence'),
    )):
        raise InvalidItemException(
            'Item cannot be instanciated from provided stickers.')

    return Item(
        song_uri,
        stickers['audiobook'],
        int(stickers['sequence']),
        float(stickers['last_position'])
        if 'last_position' in stickers else None,
        datetime.datetime.fromisoformat(stickers['last_played'])
        if 'last_played' in stickers else None,
    )


def get_item(client: musicpd.MPDClient, song_uri: str) -> Item:
    return get_item_from_sticker_map(
        song_uri, parse_stickers(client.sticker_list('song', song_uri)))


def get_items(
        client: musicpd.MPDClient,
        song_uris: typing.List[str],
) -> typing.List[Item]:
    client.command_list_ok_begin()
    for song_uri in song_uris:
        client.sticker_list('song', song_uri)
    return [
        get_item_from_sticker_map(song_uri, parse_stickers(stickers))
        for song_uri, stickers in zip(song_uris, client.command_list_end())
    ]


def get_audiobook_items(client: musicpd.MPDClient,
                        audiobook_name: str) -> typing.List[Item]:
    song_uris = [
        i['file'] for i in client.sticker_find(
            'song', '', 'audiobook', '=', audiobook_name)
    ]
    if not song_uris:
        raise SystemExit(f'No such audiobook "{audiobook_name}".')

    return sorted(get_items(client, song_uris))


def get_all_audiobooks(client: musicpd.MPDClient) -> typing.List[str]:
    audiobooks, seen = [], set()
    for item in client.sticker_find('song', '', 'audiobook'):
        audiobook = parse_stickers([item['sticker']])['audiobook']
        if audiobook in seen:
            continue
        audiobooks.append(audiobook)
        seen.add(audiobook)
    return audiobooks


def store_playback_bookmark(
        client: musicpd.MPDClient,
        item: Item,
        position: str,
        now: typing.Optional[datetime.datetime] = None,
) -> None:
    now = now or datetime.datetime.now()
    _logger.debug(
        'Storing playback bookmark for item: %s at position: %s on %s.',
        item, position, now)
    client.command_list_ok_begin()
    client.sticker_set(
        'song', item.uri, 'last_position', position,
    )
    client.sticker_set(
        'song', item.uri, 'last_played', now.isoformat(),
    )
    client.command_list_end()


def enqueue(client: musicpd.MPDClient, items: typing.List[Item],
            resume: bool = True, autoplay: bool = True) -> None:
    # If resumption is enabled, try to find the last playing position.
    last_played, last_played_idx = None, -1
    if resume:
        # Sort by last played date in descending order and take the first one
        # (latest).
        bookmarks = sorted(
            (i for i in items if i.last_played),
            key=lambda i: i.last_played,
            reverse=True,
        )
        if bookmarks:
            last_played = bookmarks[0]
            last_played_idx = items.index(last_played)

    client.command_list_ok_begin()

    # Stop current playback.
    client.stop()

    # Clear current queue.
    client.clear()

    # Turn shuffle off.
    client.random(0)

    # Turn song consumation off.
    client.consume(0)

    # Turn song repetition off.
    client.repeat(0)

    # Turn single song off.
    client.single(0)

    # Turn crossfade off.
    # TODO: Is this needed?
    # client.crossfade(0)

    # Enqueue all the items.
    for item in items:
        client.add(item.uri)

    # Seek to the last playback position (if any).
    if resume and last_played_idx >= 0:
        pos = max((0.0, (last_played.last_position or 0.0) - 5.0))
        _logger.debug(
            'Seeking last played item: %s to position: %.2f.',
            last_played, pos)
        client.seek(last_played_idx, pos)
        # `seek()` starts playback, pause it.
        client.pause()

    # Resume playback.
    if autoplay:
        client.play()

    client.command_list_end()


def do_add(args: argparse.Namespace, client: musicpd.MPDClient):
    song_uris = [ln.strip() for ln in fileinput.input(files=args.filenames)]
    client.command_list_ok_begin()
    for idx, song_uri in enumerate(song_uris, start=1):
        client.sticker_set('song', song_uri, 'audiobook', args.name)
        client.sticker_set('song', song_uri, 'sequence', str(idx))
    client.command_list_end()


def do_list(args: argparse.Namespace, client: musicpd.MPDClient):
    print('\n'.join(get_all_audiobooks(client)))


def do_continue(args: argparse.Namespace, client: musicpd.MPDClient):
    items = get_audiobook_items(client, args.name)
    enqueue(client, items, resume=True, autoplay=False)


def do_next(args: argparse.Namespace, client: musicpd.MPDClient):
    client.command_list_ok_begin()
    client.currentsong()
    client.status()
    current_song, status = client.command_list_end()
    audiobooks = get_all_audiobooks(client)

    if not audiobooks:
        raise SystemExit('There are no audiobooks.')

    current_item = None
    song_uri = current_song.get('file')
    if song_uri:
        try:
            current_item = get_item(client, song_uri)
        except InvalidItemException:
            _logger.debug(
                'Current song does not belong to an audiobook, ignoring.')

    # Find the name of the current audiobook (if any).
    next_idx = 0
    if current_item is not None:
        if len(audiobooks) == 1:
            _logger.info(
                'There is only one audiobook, no sense in switching to the '
                'next one, ignoring.')
            return

        current_idx = audiobooks.index(current_item.audiobook_name)

        # If an audiobook is playing, store its playback position before
        # switching to the next audiobook.
        position = status.get('elapsed')
        if position is not None:
            store_playback_bookmark(client, current_item, position)

        if current_idx + 1 < len(audiobooks):
            next_idx = current_idx + 1

    items = get_audiobook_items(client, audiobooks[next_idx])
    _logger.info('Enqueueing next audiobook: "%s".', audiobooks[next_idx])
    enqueue(client, items, resume=True, autoplay=True)


def do_watch(args: argparse.Namespace, client: musicpd.MPDClient):
    while True:
        events = client.idle()
        _logger.debug('Received events from MPD: %s', events)
        if 'player' in events:
            client.command_list_ok_begin()
            client.currentsong()
            client.status()
            song, status = client.command_list_end()
            song_uri = song.get('file')
            if not song_uri:
                continue

            try:
                item = get_item(client, song_uri)
            except InvalidItemException:
                _logger.debug(
                    'Current song does not belong to an audiobook, ignoring.')
                continue

            position = status.get('elapsed')
            if position:
                store_playback_bookmark(client, item, position)


def main():
    parser = argparse.ArgumentParser(
        prog='mpdab',
        description='Crude audiobook support for Music Player Daemon',
    )
    parser.add_argument(
        '--version',
        action='version',
        version=f'%(prog)s {__version__}',
    )
    parser.add_argument(
        '--debug',
        action='store_true',
    )
    parser.add_argument(
        '--host',
        help='the MPD server to connect to. This can be a hostname, IPv4/IPv6 '
             'address, an absolute path (i.e. local socket) or a name '
             'starting with @ (i.e. an abstract socket, Linux only). '
             'Default: %(default)s',
        default=os.getenv('MPD_HOST', 'localhost'),
    )
    parser.add_argument(
        '-p',
        '--port',
        type=int,
        default=os.getenv('MPD_PORT', 6600),
        help='the TCP port of the MPD server to connect to.'
             'If not given, the value of the environment variable MPD_PORT is '
             'used. Default: %(default)s',
    )
    subparsers = parser.add_subparsers()

    add_parser = subparsers.add_parser('add')
    add_parser.add_argument(
        'name',
    )
    add_parser.add_argument(
        'filenames',
        nargs='+',
        default=['-'],
    )
    add_parser.set_defaults(func=do_add)

    list_parser = subparsers.add_parser('list', aliases=('ls',))
    list_parser.set_defaults(func=do_list)

    continue_parser = subparsers.add_parser('continue')
    continue_parser.add_argument(
        'name',
    )
    continue_parser.set_defaults(func=do_continue)

    next_parser = subparsers.add_parser('next')
    next_parser.set_defaults(func=do_next)

    watch_parser = subparsers.add_parser('watch')
    watch_parser.set_defaults(func=do_watch)

    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    client = get_client_from_args(args)

    client.connect()
    if client.pwd:
        client.password(client.pwd)

    func = getattr(args, 'func', lambda *a: parser.print_usage())
    try:
        func(args, client)
    except KeyboardInterrupt:
        print('Interrupted by user, exiting...')


if __name__ == '__main__':
    main()
